from PyDSTool.common import *
from PyDSTool.errors import *
from PyDSTool.utils import remain, info
from PyDSTool.Points import Point, Pointset
from numpy import array, asarray, NaN, Inf, isfinite
from PyDSTool import figure, plot

_classes = ['connection', 'node', 'simulator', 'composed_map1D',
            'map', 'map1D', 'map2D', 'identity_map', 'delay_map']

_functions = ['extract_digraph']

_instances = ['idmap']

__all__ = _classes + _functions + _instances

# ------------------------------------------------------------------------------

class connection(object):
    def __init__(self, name):
        self.name = name
        # input is a node object
        self.input = None
        # output is for informational purposes only
        # (nodes keyed by name)
        self.outputs = {}
        # map is of type map (defaults to identity)
        self.map = idmap

    def poll(self):
        # poll input
        return self.map(self.input.last_t)

    def __repr__(self):
        return "connection(%s)" % self.name

    __str__ = __repr__


class node(object):
    def __init__(self, name):
        self.name = name
        # inputs are connection objects
        self.inputs = {}
        # output is for informational purposes only
        # (connections keyed by name)
        self.outputs = {}
        # projected state
        self.next_t = 0
        # last state
        self.last_t = 0
        # current input values
        self.in_vals = {}
        # map is of type map
        self.map = None
        # FIFO (sorted append) queue for future events
        self.resetQ()

    def poll(self, state):
        # poll inputs
        self.in_vals.update(state)
        # overwrite current known state with projected values from inputs
        for name, connxn in self.inputs.items():
            self.in_vals[connxn.input.name] = connxn.poll()
        self.next_t = self.map(self.in_vals)
        return self.next_t

    def addToQ(self, item):
        self.queue.append(item)
        self.queue.sort()

    def resetQ(self, seq=None):
        if seq is None:
            self.queue = []
        else:
            # ensure list argument (for pop to work)
            self.queue = list(seq)

    def popFromQ(self):
        return self.queue.pop(0)

    def __repr__(self):
        return "node(%s)" % self.name

    __str__ = __repr__



def extract_digraph(mspec, node_types, connxn_types):
    connxn_names = mspec.search(connxn_types)
    node_names = mspec.search(node_types)

    # declare connection and node objects
    connxns = {}
    nodes = {}
    for c in connxn_names:
        cobj = mspec.components[c]
        new_connxn = connection(c)
        connxns[c] = new_connxn

    for n in node_names:
        nobj = mspec.components[n]
        new_node = node(n)
        nodes[n] = new_node

    # fill in inputs and outputs dictionaries for each type
    for cn, c in connxns.items():
        cobj = mspec.components[cn]
        targs = [head(t) for t in cobj.connxnTargets]
        c.outputs = dict(zip(targs, [nodes[t] for t in targs]))
        for t in targs:
            nodes[t].inputs[cn] = c

    for nn, n in nodes.items():
        nobj = mspec.components[nn]
        targs = [head(t) for t in nobj.connxnTargets]
        n.outputs = dict(zip(targs, [connxns[t] for t in targs]))
        for t in targs:
            connxns[t].input = n

    return nodes, connxns


def head(hier_name):
    if '.' in hier_name:
        return hier_name.split('.')[0]
    else:
        return hier_name

def tail(hier_name):
    if '.' in hier_name:
        return hier_name.split('.')[-1]
    else:
        return hier_name

# maps' inputs and output are absolute times --
# for dealing with relative times they must be passed a reference
# time value to add to a relative time.
class map(object):
    pass

class map2D(map):
    pass

class map1D(map):
    pass


class delay_map(map1D):
    def __init__(self, delay):
        self.delay = delay

    def __call__(self, t):
        return t + self.delay

class identity_map(map1D):
    def __call__(self, t):
        return t

# default instance of identity class
idmap = identity_map()


class composed_map1D(map1D):
    def __init__(self, m1, m2):
        self.m1 = m1
        self.m2 = m2

    def __call__(self, t):
        return self.m1(self.m2(t))

# -----------------------------------------------------------------------------

class simulator(object):
    def __init__(self, nodes, connections, state=None):
        self.nodes = nodes
        # don't really need the connections, but just as a reference
        self.connections = connections
        if state is None:
            self.state = dict(zip(nodes.keys(), [NaN for n in nodes]))
        else:
            self.state = state
        self.curr_t = 0
        self.history = {}
        self.verbosity = 0

    def set_state(self):
        for name, n in self.nodes.items():
            n.last_t = self.state[name]

    def validate(self):
        for name, n in self.nodes.items():
            assert isinstance(n, node)
        for name, connxn in self.connections.items():
            assert isinstance(connxn, connection)
        assert remain(self.state.keys(), self.nodes.keys()) == []

    def run(self, ics, t_end):
        self.curr_t = max(ics.values())
        self.state = ics
        self.validate()
        node_names = sortedDictKeys(self.state)
        self.history = {}

        while self.curr_t <= t_end:
            self.set_state()
            if self.verbosity > 0:
                try:
                    self.display()
                except PyDSTool_BoundsError:
                    print "Maps borked at", self.curr_t
                    break
            self.history[self.curr_t] = self.state.copy()
            try:
                self.next_state = self.compile_next_state()
            except PyDSTool_BoundsError:
                print "Maps borked at", self.curr_t
                break
            vals = sortedDictValues(self.next_state)
            filt_vals = []
            min_val = Inf
            min_ix = None
            for i, v in enumerate(vals):
                # ignore projected times that are in the past now
                if v > self.curr_t:
                    if v < min_val:
                        min_val = v
                        min_ix = i
            if min_ix is None:
                # no further events possible
                print "no further events possible, stopping!"
                break
            next_node = node_names[min_ix]
            self.curr_t = vals[min_ix]
            if self.verbosity > 0:
                print "Next node is", next_node, "at time ", self.curr_t
            self.state[next_node] = self.curr_t

        ts, val_dicts = sortedDictLists(self.history, byvalue=False)
        vals = []
        for vd in val_dicts:
            vals.append([vd[nname] for nname in node_names])
        self.result = Pointset(indepvararray=ts, indepvarname='t',
                               coordnames = node_names,
                               coordarray = array(vals).T)

    def compile_next_state(self):
        vals = {}
        for name, n in self.nodes.items():
            vals[name] = n.poll(self.state)
        return vals

    def display(self):
        print "\n\n****** t =", self.curr_t
        info(self.state, "known state")
        print "\nNodes:"
        for name, n in self.nodes.items():
            n.poll(self.state)
            print name
            for in_name, in_val in n.in_vals.items():
                print "  Input", in_name, ": ", in_val

    def extract_history_events(self):
        node_names = self.nodes.keys()
        node_events = dict(zip(node_names,[None]*len(node_names)))
        ts = sortedDictKeys(self.history)
        old_state = self.history[ts[0]]
        for nn in node_names:
            node_events[nn] = [old_state[nn]]
        for t in ts[1:]:
            state = self.history[t]
            for nn in node_names:
                if old_state[nn] != state[nn]:
                    node_events[nn].append(state[nn])
            old_state = state
        return node_events

    def display_raster(self):
        h = self.history
        ts = sortedDictKeys(h)
        node_names = self.nodes.keys()
        print "\n\nNode order in plot (bottom to top) is", node_names
        figure()
        t0 = ts[0]
        state = h[t0]
        # show all i.c.s
        for ni, n in enumerate(node_names):
            plot(state[n], ni, 'ko')
        for t in ts:
            state = h[t]
            for ni, n in enumerate(node_names):
                if state[n] == t:
                    plot(t, ni, 'ko')


def sequences_to_eventlist(seq_dict):
    """seq_dict maps string symbols to increasing-ordered sequences of times.
    Returns a single list of (symbol, time) pairs ordered by time."""
    out_seq = []
    symbs = seq_dict.keys()
    next_s = None
    indices = {}
    for s in symbs:
        indices[s] = 0
    remaining_symbs = symbs
    while remaining_symbs != []:
        #print "\n***", remaining_symbs
        to_remove = []
        #print indices
        #print out_seq, "\n"
        next_t = Inf
        for s in remaining_symbs:
            try:
                t = seq_dict[s][indices[s]]
            except IndexError:
                # no times remaining for this symbol
                #print "No more symbols for ", s
                to_remove.append(s)
            else:
                #print s, t
                if t < next_t:
                    next_s = s
                    next_t = t
                    #print "Chose ", next_s, t
        indices[next_s] += 1
        for s in to_remove:
            remaining_symbs.remove(s)
        if isfinite(next_t):
            out_seq.append( (next_s, next_t) )
    return out_seq

# ---------------------------------------------------------------------

if __name__ == '__main__':
    class testmap1D(map1D):
        def __init__(self, targ, val):
            self.targ = targ
            self.val = val

        def __call__(self, state):
            return self.val + state[self.targ]

    class testmap2D(map2D):
        def __init__(self, targ, me):
            self.targ = targ
            self.me = me

        def __call__(self, state):
            return 3 + state[self.targ] + 0.1*state[self.me]

    xmap = testmap1D('x', 2)
    ymap = testmap2D('y', 'x')
    x = node('x')
    y = node('y')

    xcy = connection('xcy')
    xcy.input = x
    xcy.outputs = {'y': y}

    ycx = connection('ycx')
    ycx.input = y
    ycx.map = delay_map(2)
    ycx.outputs = {'x': x}

    xcx = connection('xcx')
    xcx.input = x
    xcx.outputs = {'x': x}

    x.inputs['y'] = ycx
    x.inputs['x'] = xcx
    x.map = ymap
    y.inputs['x'] = xcy
    y.map = xmap

    sim = simulator({'x': x, 'y': y}, {'xcy': xcy, 'ycx': ycx,
                                       'xcx': xcx})
    sim.verbosity = 1
    sim.run({'x': 0, 'y': -2}, 10)
